#ifndef _CRPFILE_H
#define _CRPFILE_H

// Internal representation of CRP file fragments
// (C) 2003 Frank Paehlke <frank@paehlke.de>
// $Id: crpfile.h,v 1.3 2004/01/29 09:31:11 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <string>
#include <vector>
#include <ctime>

#include "smooth.h"

using namespace std;

//! Internal representation of a CicloTour CRP file

class CRPFile {
 public:
        
    //! Exception thrown when a CRP file to be read contains errors
    
    class MalformedInputException {};
    
    // constructors
    CRPFile (smooth* _sm = 0) : sm (_sm) {}
    CRPFile (const CRPFile&);

    // data manipulation methods

    void clear();

    // get aggregated data values
    
    double get_hdiff() const throw() { return hdiff; }
    double get_hdiff_up() const throw() { return hdiff_up; }
    double get_hdiff_down() const throw() { return hdiff_down; }
    double get_length() const throw() { return length; }
    double get_length_up() const throw() { return length_up; }
    double get_length_down() const throw() { return length_down; }
    double get_avgslope() const throw() { return avgslope; }
    double get_avgslope_up() const throw() { return avgslope_up; }
    double get_avgslope_down() const throw() { return avgslope_down; }
    double get_maxslope() const throw() { return maxslope; }
    double get_minslope() const throw() { return minslope; }

    // get/set meta data (title, date etc.)

    void set_title(const string& t) { title = t; }
    string get_title() const { return title; }

    tm get_date() const { return tourdate; }

    //! set smoothing algorithm

    void set_smooth (smooth * _sm) throw() { sm = _sm; }

    // input/output of CRP files

    friend istream& operator >> (istream&, CRPFile&)
	throw (MalformedInputException);

    friend ostream& operator << (ostream&, const CRPFile&);

    // write data as CSV file

    void writeCSV (ostream&) const;

    // write meta data as XML file

    void writeXML (ostream&) const;

 private:

    // calculate derived data rows and aggregated values
    
    void calculate();

    // raw data read from file:
    vector<string> time; // time [hh:mm:ss]
    vector<double> distance; // distance [km]
    vector<double> height; // height [m]
    vector<int> hrate; // heart rate [1/min]
    vector<double> temperature; // temperature [C]
    vector<string> comment; // user comments
    
    // data rows computed from raw data
    vector<double> cdistance; // corrected distance [km]
    vector<double> cheight; // corrected height [m]
    vector<double> speed; // speed [km/h]
    vector<double> vspeed; // vertical speed [m/h]
    vector<double> slope; // slope [%]
    
    // aggregated data
    int nrows; // number of table rows
    double length, length_up, length_down;
    double hdiff, hdiff_up, hdiff_down;
    double avgslope, avgslope_up, avgslope_down;
    double maxslope, minslope;

    // meta data
    string title;	//! Title of the tour
    tm tourdate;	//! Date when the tour was made

    smooth * sm;	//! smoothing algorithm
};

#if 0

//! One data row in a CRP file
 
class CRPRow {
 public:   
    // raw data read from file

    string time;	//! time [hh:mm:ss]
    double distance;	//! distance [km]
    double height;	//! height [m]
    int hrate;		//! heart rate [1/min]
    double temperature; //! temperature [C]
    string comment;	//! user comment
    
    // computed values

    double cdistance;	//! corrected distance [km]
    double cheight;	//! corrected height [m]
    double speed;	//! avg. speed of last interval [km/h]
    double vspeed;	//! avg. vertical speed of last interval [m/h]
    double slope;	//! avg. slope of last interval [%]
};

#endif

#endif
