#! /usr/bin/env python

# Create an SVG graph from a CSV file with height profile data.
# The CSV file must contain the distance and height values in
# the 6th and 7th column, respectively.
# CSV files which fulfill this condition are created by crp2csv.

# Author: Frank Paehlke
# $Id: svggraph.py,v 1.4 2004/09/29 19:16:25 frank Exp $

import sys
import math
from optparse import OptionParser
import string
import csv
import array
import os
import tempfile
import crputils_version

def main():
    global options
    # parse command line arguments
    parser = OptionParser(usage="%prog [options] [csvfile [svgfile]]",
                          version=crputils_version.version)
    parser.add_option ("-v", "--verbose", action="store_true",
                       dest="verbose", default=False,
                       help="turn on debugging output")
    parser.add_option ("-t", "--title", dest="title", default="", 
                       help="set title text")
    parser.add_option ("-c", "--copyright", dest="copyright", default="",
                       help="set copyright text")
    parser.add_option ("-f", "--format", dest="format", default="svg",
                       help="output format: svg (default), "+
                       "png (requires ImageMagick and Netpbm), "+
                       "pdf (requires batik-rasterizer)")
    parser.add_option ("", "--dscale", dest="dscale", default="",
                       metavar="MIN,MAX",
                       help="set scale of distance axis manually" )
    parser.add_option ("", "--hscale", dest="hscale", default="",
                       metavar="MIN,MAX",
                       help="set scale of height axis manually")
    parser.add_option ("", "--sscale", dest="sscale", default="",
                       metavar="MIN,MAX",
                       help="set scale of slope axis manually")
    (options, args) = parser.parse_args()

    # do some checks
    options.format = string.lower(options.format)
    if options.format not in ['svg','png','pdf']:
        print >> sys.stderr, "unsupported format:", options.format
        sys.exit(1)
    if len(args) > 2:
        print >> sys.stderr, "Ignoring extra argument(s):", args[2:]

    # read input/output file name
    if len(args) >= 1:
        ifilename = args[0]
    else:
        ifilename = "-"
    if len(args) >= 2:
        ofilename = args[1]
    else:
        i = string.rfind(ifilename,'.')
        if i < 0:
            basename = ifilename
        else:
            basename = ifilename[0:i]
        ofilename = basename+'.'+options.format
    if options.verbose:
        print >> sys.stderr, "Input file:", ifilename
        print >> sys.stderr, "Output file:", ofilename
        print >> sys.stderr, "Output format:", options.format
        
    # read input file
    if ifilename == '-':
        ifile = sys.stdin
    else:
        ifile = open(ifilename)
    dist,height = readFile(ifile)
    if (ifile != sys.stdin):
        ifile.close()

    if options.verbose:
        print >> sys.stderr, "Title:", options.title
        print >> sys.stderr, "Copyright text:", options.copyright

    if options.format == "svg":
        # write output file
        if ofilename == '-':
            ofile = sys.stdout
        else:
            ofile = open(ofilename, "w")
        printGraph(ofile, dist, height, options.format)
        if ofile != sys.stdout:
            ofile.close()

    elif options.format == "png":
        # create web color map using Netpbm
        tfd,tfname = tempfile.mkstemp(".pnm")
        if options.verbose:
            print >> sys.stderr, "Temporary file for color map:", tfname
        os.close(tfd)
        os.system("pamseq -tupletype RGB 3 5 | pamtopnm > " + tfname)
        pipe = os.popen("convert -map " + tfname
                        + " +dither -depth 8 -quality 90"
                        + " - " + ofilename, "w");
        printGraph(pipe, dist, height, options.format)
        pipe.close()
        os.remove(tfname)

    elif options.format == "pdf":
        tfd,tfname = tempfile.mkstemp(".svg")
        if options.verbose:
            print >> sys.stderr, "Temporary file:", tfname
        ofile = os.fdopen (tfd,'w')
        printGraph(ofile, dist, height, options.format)
        ofile.close()
        os.spawnlp(os.P_WAIT, 'batik-rasterizer', 'batik-rasterizer',
                   '-m', 'application/pdf', '-d', ofilename, tfname )
        os.remove(tfname)

# read input data from file f
# returns: dist[], height[]

def readFile(f):
    dist = array.array('d')
    height = array.array('d')
    reader = csv.reader(f,delimiter='\t')
    for line in reader:
        # does the line contain a title?
        token = line[0].strip().lower()
        if token == "title":
            if options.title == "":
                options.title = line[1]
        elif token == "copyright":
            if options.copyright == "":
                options.copyright = line[1]
        elif token == "dscale":
            if options.dscale == "":
                options.dscale = line[1].strip()
        elif token == "hscale":
            if options.hscale == "":
                options.hscale = line[1].strip()
        elif token == "sscale":
            if options.sscale == "":
                options.sscale = line[1].strip()
        elif line[0][0] in "0123456789.+-":
            # read distance and height
            dist.append(float(line[5]))
            height.append(float(line[6]))
    if options.verbose:
        print >> sys.stderr, len(dist), "data rows read"
    return dist,height
        
def printGraph(f,dist,height,outputFormat):
    printHeader(f,outputFormat)
    # calculate slope
    slope = map((lambda a,b,c,d: (b-a)/(d-c)/10.0),
                height[0:-1], height[1:], dist[0:-1], dist[1:])
    # calculate scales
    d0,d1,dIntv = calcXScale(dist)
    h0,h1,hIntv,s0,s1,sIntv = calcYScale(height,slope)
    dTics = int((d1-d0)/dIntv)
    hTics = (h1-h0)/hIntv
    sTics = (s1-s0)/sIntv
    # background
    f.write('<rect x="0" y="0" width="639" height="299" fill="#ffffff" stroke="#000000" stroke-width="1" />\n')

    # title
    if options.title != "":
        f.write('<text x="320" y="30" text-anchor="middle"')
        f.write(' font-weight="bold" font-size="16">\n')
        f.write(options.title + '\n')
        f.write('</text>\n')

    # copyright
    if options.copyright != "":
       f.write('<text x="636" y="296" text-anchor="end" font-size="10" fill="#999999">')
       f.write(options.copyright);
       f.write('</text>\n')

    # canvas
    f.write('<g style="font-size:14" transform="translate(80,50)">\n')
    f.write(' <rect x="0" y="0" width="500" height="200" fill="#cccccc" />\n')

    # X minor grid
    if dTics <= 12:
        f.write(' <g style="stroke:#999999;stroke-width:1">\n')
        for i in range (dTics):
            d = d0 + (i+0.5)*dIntv;
            x = int(round(scale(d, d0, d1, 0, 500)))
            f.write('  <line x1="' + `x` + '" y1="0" x2="' + `x` + '" y2="200" />\n');
        f.write(' </g>\n')

    # Y minor grid
    if sTics <= 7:
        f.write(' <g style="stroke:#999999;stroke-width:1">\n')
        for i in range (sTics):
            s = s0 + (i+0.5)*sIntv;
            y = int(round(scale(s, s0, s1, 200, 0)))
            f.write('  <line x1="0" y1="' + `y` + '" x2="500" y2="' + `y` + '" />\n');
        f.write(' </g>\n')

    # X grid
    f.write(' <g style="stroke:#000000;stroke-width:1">\n')
    for i in range (dTics+1):
        d = d0 + i*dIntv;
        x = int(round(scale(d, d0, d1, 0, 500)))
        f.write('  <line x1="' + `x` + '" y1="0" x2="' + `x` + '" y2="205" />\n');
    f.write(' </g>\n')

    # X labels
    f.write(' <g style="text-anchor:middle">\n')
    f.write('  <text x="250" y="240">Strecke [km]</text>\n')
    for i in range (dTics+1):
        d = d0 + i*dIntv;
        x = int(round(500 * float(d-d0) / float(d1-d0)))
        f.write('  <text x="' + `x` + '" y="220">'
                + str(round(d,1)) + '</text>\n');
    f.write(' </g>\n')

    # Y grid
    f.write(' <g style="stroke:#000000;stroke-width:1">')
    for i in range (hTics+1):
        h = h0 + i*hIntv;
        s = s0 + i*sIntv;
        y = round(scale(h, h0, h1, 200, 0))
        f.write('  <line x1="-5" y1="' + `y` + '" x2="505" y2="' + `y` + '" ');
        if (s == 0) or (h == 0): # zero axis
            f.write('style="stroke-width:2" ')
        f.write('/>\n')
    f.write(' </g>\n')

    # Y labels (primary)
    f.write(' <g transform="translate(-55,100) rotate(-90)">\n')
    f.write('  <text x="0" y="0" text-anchor="middle">blau: H&#246;he [m]</text>\n')
    f.write(' </g>\n')
    f.write(' <g style="text-anchor:end">\n')
    for i in range (hTics+1):
        h = h0 + i*hIntv;
        y = round(scale(h, h0, h1, 200, 0))
        f.write('  <text x="-10" y="' + `y` + '" dy="0.5ex">'
                + str(int(round(h))) + '</text>\n');
    f.write(' </g>')

    # Y labels (secondary)
    f.write(' <g transform="translate(545,100) rotate(-90)">\n')
    f.write('  <text x="0" y="0" text-anchor="middle">rot: Steigung [%]</text>\n')
    f.write(' </g>\n')
    f.write(' <g style="text-anchor:start">\n')
    for i in range (sTics+1):
        s = s0 + i*sIntv
        y = round(scale(s, s0, s1, 200, 0))
        f.write('  <text x="510" y="' + `y` + '" dy="0.5ex">'
                + str(int(round(s))) + '</text>\n');
    f.write(' </g>\n');

    # slope curve (first, i.e. behind height curve)
    dist2 = map((lambda x,y: (x+y)/2), dist[0:-1], dist[1:])
    printCurve(f,dist2,d0,d1,slope,s0,s1,"#ff0000")

    # height curve (second, i.e. before slope curve)
    printCurve(f,dist,d0,d1,height,h0,h1,"#0000ff")

    f.write('</g>\n')
    printTrailer(f, outputFormat)
    return

# print SVG file header

def printHeader(f,outputFormat):
    f.write('<?xml version="1.0" encoding="iso-8859-1"?>\n')
    f.write('<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.0//EN" "http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd">')
    if outputFormat=="png":
        f.write('<svg width="640px" height="300px" version="1.0" viewBox="0 0 640 300"\n')
        f.write(' preserveAspectRatio="none" xmlns="http://www.w3.org/2000/svg">\n')
        f.write('<g style="font-family:Arial,Helvetica,sans-serif">\n')
    elif outputFormat=="pdf":
        # page size A4 which centered viewBox (640,300)x0.4mm
        f.write('<svg width="297mm" height="210mm" version="1.0" viewBox="-51.25 -112.5 742.5 525"\n')
        f.write(' preserveAspectRatio="none" xmlns="http://www.w3.org/2000/svg">\n')
        f.write('<g style="font-family:Arial,Helvetica,sans-serif">\n')
    else:
#        f.write('<svg width="'+str(640.0/72)+'in" height="'
#        +str(300.0/72)+'in" version="1.0" viewBox="0 0 640 300"\n')
        f.write('<svg width="640pt" height="300pt" version="1.0" viewBox="0 0 640 300"\n')
        f.write(' preserveAspectRatio="none" xmlns="http://www.w3.org/2000/svg">\n')
        f.write('<g style="font-family:Arial,Helvetica,sans-serif" transform="translate(0.5,0.5)">\n')

# print SVG file trailer

def printTrailer(f,outputFormat):
    f.write('</g>\n')
    f.write('</svg>\n')
    
# calculate scaling parameters X axis
# returns: dBottom, dTop, dIntv

def calcXScale (dist):
    if options.dscale != "":
        (dMin, dMax) = string.split(options.dscale, ',')
        dMin = float(dMin)
        dMax = float(dMax)
    else:
        dMin = min(dist)
        dMax = max(dist)
    if options.verbose:
        print "D min/max:", dMin, dMax

    # determine scale
    for dIntv in [0.1, 0.2, 0.5, 1, 2, 4, 5]:
        # round up/down
        dBottom = dIntv * math.floor(dMin/dIntv)
        dTop = dIntv * math.ceil(dMax/dIntv)
        dTics = int(round((dTop-dBottom) / dIntv))
        if (dTics <= 15):
            break # we have found a suitable scale
    if options.verbose:
        print "D scale:", dBottom, dTop, dIntv, dTics
    return dBottom, dTop, dIntv

# calculate scaling parameters for primary and secondary Y axis
# returns: hBottom, hTop, hIntv, sBottom, sTop, sIntv

def calcYScale (height,slope):
    if options.hscale != "":
        (hMin, hMax) = map(float, string.split(options.hscale, ','))
    else:
        hMin = min(height)
        hMax = max(height)
    if options.sscale != "":
        (sMin, sMax) = map(float, string.split(options.sscale, ','))
    else:
        # always include zero axis into slope scale
        sMin = min(slope+[0])
        sMax = max(slope+[0])
    # determine scale for primary axis only
    for hIntv in [5, 10, 20, 50, 100, 200, 400]:
        # round up/down
        hBottom = hIntv*int(math.floor(hMin/hIntv))
        hTop = hIntv*int(math.ceil(hMax/hIntv))
        # hBottom and hTop are integer!
        hTics = (hTop-hBottom) / hIntv
        if (hTics <= 8):
            break # we have found a suitable scale
    if options.verbose:
        print "original H scale:", hBottom, hTop, hIntv, hTics

    # determine scale for secondary axis only
    for sIntv in [1, 2, 4]:
        # round up/down
        sBottom = sIntv*int(math.floor(sMin/sIntv))
        sTop = sIntv*int(math.ceil(sMax/sIntv))
        # sBottom and sTop are integer!
        sTics = (sTop-sBottom) / sIntv
        if (sTics <= 8):
            break # we have found a suitable scale
    if options.verbose:
        print "original S scale:", sBottom, sTop, sIntv, sTics

    # adjust scales to match each other
    if (sTics > hTics):
        # adjust height scale
        d = sTics - hTics
        hBottom -= hIntv * (d/2)
        hTop += hIntv * (d/2)
        if d%2 == 1: # odd number
            if (hTop-hMax < hMin-hBottom):
                hTop += hIntv
            else:
                hBottom -= hIntv
        hTics = sTics
    elif (hTics > sTics):
        # adjust slope scale
        d = hTics - sTics
        if sBottom == 0:
            sTop += sIntv * d
        else:
            sBottom -= sIntv * (d/2)
            sTop += sIntv * (d/2)
            if d%2 == 1: # odd number
                if (sTop-sMax < sMin-sBottom):
                    sTop += sIntv
                else:
                    sBottom -= sIntv
        sTics = hTics
    if options.verbose:
        print "adjusted H scale:", hBottom, hTop, hIntv, hTics
        print "adjusted S scale:", sBottom, sTop, sIntv, sTics
    return hBottom, hTop, hIntv, sBottom, sTop, sIntv

# print actual graph

def printCurve(f,x,xMin,xMax,y,yMin,yMax,color):
    # calculate derivatives for Bezier points
    x1 = derive(x)
    y1 = derive(y)
    # begin SVG path statement
    f.write('<path style="stroke:' + color + ';stroke-width:1.5;fill:none" ')
    f.write('d="');
    # first point
    sx,sy = scaleCanvas (x[0], xMin, xMax, y[0], yMin, yMax)
    f.write('M' + str(round(sx,2)) + ',' + str(round(sy,2)))
    # remaining points
    for i in range(1,len(x)):
        # bezier segment: first control point
        sx,sy = scaleCanvas (x[i-1] + x1[i-1]/3.0, xMin, xMax,
                             y[i-1] + y1[i-1]/3.0, yMin, yMax)
        f.write(' C' + str(round(sx,2)) + ',' + str(round(sy,2)))
        # second control point
        sx,sy = scaleCanvas (x[i] - x1[i]/3.0, xMin, xMax,
                             y[i] - y1[i]/3.0, yMin, yMax)
        f.write(',' + str(round(sx,2)) + ',' + str(round(sy,2)))
        # end point
        sx,sy = scaleCanvas (x[i], xMin, xMax, y[i], yMin, yMax)
        f.write(',' + str(round(sx,2)) + ',' + str(round(sy,2)))
    f.write('" />\n')
        
# scale coordinate such that the interval [x0,x1] is mapped
# to [u0,u1]

def scale(x,x0,x1,u0,u1):
    return u0+(u1-u0)*float(x-x0)/(x1-x0)

# scale point to fit SVG canvas (0,0)-(500,200)

def scaleCanvas(x,x0,x1,y,y0,y1):
    sx = scale(x,x0,x1,0,500)
    sy = scale(y,y0,y1,200,0)
    return sx,sy

# approximate derivative of an array of data points
# using quadratic interpolation between consecutive points

def derive(d):
    n = len(d) # number of points
    if n <= 1:
        return d
    elif n == 2:
        return [d[1]-d[0], d[1]-d[0]]
    else:
        d1 = array.array('d') # result
    
        # calculate d1[0] from quadratic interpolation of d[0..2]
        d1.append((4*d[1] - 3*d[0] - d[2]) / 2.0)
        for i in range(1, n-1):
            # calculate d1[i] from adjacent points
            if (d[i+1] - d[i] >= 0 and d[i]-d[i-1] <= 0) \
               or (d[i+1] - d[i] <= 0 and d[i]-d[i-1] >= 0):
                di = 0.0
            else:
                di = (d[i+1] - d[i-1]) / 2.0
            d1.append(di)
        # calculate d1[n-1] from quadratic interpolation of d[n-3..n-1]
        d1.append((3*d[n-1] + d[n-3] - 4*d[n-2]) / 2.0)
        return d1

main()
